"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const fs = require('fs');
/// <reference path="../types/left-pad/index.d.ts" />
const leftPad = require('left-pad');
/**
 * Find the related schema when referenced from a property
 * Throws error if schema could not be found
 * @param arrSchemas {Array<Schema>} the array of schema's we have imported to search through
 * @param relatedId {String} the ID for the schema we need to find
 * @returns {Schema} The referenced schema
 */
function getRelatedSchema(arrSchemas, relatedId) {
    // find schema by ID
    const relatedSchema = arrSchemas.find(s => s.id === relatedId);
    if (!relatedSchema) {
        throw new Error(`Could not find referenced schema "${relatedId}"`);
    }
    return relatedSchema;
}
/**
 * Sort the properties in the outputed interface by key alphabetically (case insensitive)
 * @param a {Property} the first property to compare
 * @param b {Property} the second property to compare
 * @returns {number} 0 if equal, 1 if a should come after b, -1 if b should come after a
 */
function sortProperties(a, b) {
    // case insensitive sort
    const nameA = a.key.toUpperCase();
    const nameB = b.key.toUpperCase();
    if (nameA < nameB) {
        return -1;
    }
    if (nameA > nameB) {
        return 1;
    }
    return 0;
}
function sortImports(a, b) {
    const nameA = `./${a.outputFileName}`;
    const nameB = `./${b.outputFileName}`;
    if (nameA < nameB) {
        return -1;
    }
    if (nameA > nameB) {
        return 1;
    }
    return 0;
}
/**
 * Create module import statements for any referenced schemas (if applicable)
 * NOTE imports are only created when the referenced schema is of type object.
 * @param schema {Schema} the schema to generate imports for
 * @param arrSchemas {Array<Schema>} the array of schemas to search through to find referenced schemas
 * @returns {string} a string of imports for the supplied schemas (empty string if no imports required)
 */
function writeImports(schema, arrSchemas) {
    if (schema.dependencies.length < 1) {
        return '';
    }
    let deps = schema.dependencies
        .map(dep => getRelatedSchema(arrSchemas, dep)) // find the related schema
        .filter(s => (s.type === 'object' || s.type === 'array') && s.hasProperties()) // only import if schema is of type object
        .sort(sortImports)
        .map(s => `import {${s.title}} from './${s.outputFileName}';`); // generate ES6 module import
    // if we filtered out all the imports, rather than leaving blank lines return empty string
    if (deps.length < 1) {
        return '';
    }
    // get uniques
    const temp = deps;
    deps = [];
    // remove any duplicates by copying between arrays
    temp.forEach((item) => {
        if (deps.indexOf(item) === -1) {
            deps.push(item);
        }
    });
    return deps.join(('\n')) + '\n\n'; // write each import to a new line, leave 2 lines between imports and the interface definition
}
/**
 * Write each one of the properties and it's relevant type as detailed in the schema to the interface file recursively
 * @param schema {Schema} the schema to get the properties from
 * @param arrSchemas {Array<Schema>} the array of all known schemas used to lookup $ref properties by id
 * @param indentation {number} the number of spaces to pad from the left (for style only)
 * @returns {string} the properties of the schema with their associated typescript types
 */
function writeProperties(schema, arrSchemas, indentation = 2) {
    return schema.properties
        .sort(sortProperties) // sort properties alphabetically
        .map(property => {
        // get the key and sort the indentation out
        let result = `${property.key}: `;
        result = leftPad(result, result.length + indentation);
        switch (property.typescriptType) {
            case '___OBJECT___':
                // if the property is an object, create a child object using it's properties (and indent them an extra 2 spaces)
                result += `{\n${writeProperties(property, arrSchemas, (indentation + 2))}\n  };`;
                break;
            case '___REFERENCE___':
                // if the property references another schema, find the related schema (this will throw if not found so no need to check)
                const related = getRelatedSchema(arrSchemas, property.ref);
                if (related.type === 'object') {
                    let temp;
                    if (related.hasProperties()) {
                        // if the referenced item is another object - reference it's interface
                        temp = `${getRelatedSchema(arrSchemas, property.ref).title}`;
                    }
                    else {
                        temp = 'any';
                    }
                    // if type is array, wrap it up in an array Object, otherwise return the interface name
                    if (property.isArray()) {
                        result += `Array<${temp}>;`;
                    }
                    else {
                        result += `${temp};`;
                    }
                }
                else {
                    // if we get a property which isn't an object, convert it to a property and get it's type
                    const prop = related.toProperty();
                    result += `${prop.typescriptType};`;
                }
                break;
            default:
                // by default write the property type
                result += `${property.typescriptType};`;
        }
        // if property is not required add the ? after the key
        if (!schema.isPropertyRequired(property.key)) {
            result = result.replace(':', '?:');
        }
        return result;
    })
        .join('\n'); // write each property on it's own line
}
/**
 * Do the writing to disk
 * @param template {string} the generated interface to write to disk
 * @param schema {Schema} the schema we're writing (used to get the file name)
 * @param outputDirectory {string} the directory to write the files to (defaults to current directory if not specified)
 * @returns {Promise<string>} a promise which completes when the file has been written returning the file path of the created file
 */
function writeToFile(template, schema, outputDirectory = './') {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            const fileName = `${outputDirectory}/${schema.outputFileName}.d.ts`;
            fs.writeFile(fileName, template, (err) => {
                return err ? reject(err) : resolve(fileName);
            });
        });
    });
}
/**
 * Generate the TypeScript interface file and write it to disk
 * @param schema {Schema} the schema to generate the interface for
 * @param arrSchemas {Array<Schema>} the array of all schemas we know about (used to look up related schemas)
 * @param outputDirectory {String} the directory to write the files to (defaults to current directory if not specified)
 * @returns {Promise<string>} a promise which completes when the file has been written returning the file path of the created file
 */
function generateInterfaceFile(schema, arrSchemas, outputDirectory) {
    return __awaiter(this, void 0, void 0, function* () {
        if (schema.type !== 'object') {
            // if the schema is not an object schema we can't create an interface for it, just log and return
            console.warn(schema.title, 'is not of type object so no interface will be created');
            return Promise.resolve('');
        }
        if (schema.properties.length < 1) {
            console.warn(schema.title, 'has no child properties, using "any" as the type instead of creating interface');
            return Promise.resolve('');
        }
        // generate the template and write it (this is all on 1 line to avoid extra spaces)
        const template = `${writeImports(schema, arrSchemas)}export interface ${schema.title} {\n${writeProperties(schema, arrSchemas)}\n}\n`;
        return yield writeToFile(template, schema, outputDirectory);
    });
}
exports.generateInterfaceFile = generateInterfaceFile;
